@@ -3,12 +3,14 @@ import os
 import sys
 import typing
 
+import async_generator
 import attr
 from qtpy import QtWidgets
 import trio
 
 import qtrio
 import qtrio._qt
+import qtrio._util
 
 
 if sys.version_info >= (3, 8):
@@ -17,13 +19,12 @@ else:
     from typing_extensions import Protocol
 
 
-class DialogProtocol(Protocol):
-    """The common interface used for working with QTrio dialogs.  To check that a class
-    implements this protocol see :func:`qtrio.dialogs.check_dialog_protocol`.
+class BasicDialogProtocol(Protocol):
+    """The minimal common interface used for working with QTrio dialogs.  To check that
+    a class implements this protocol see
+    :func:`qtrio.dialogs.check_basic_dialog_protocol`.
     """
 
-    shown: qtrio.Signal
-    """The signal to be emitted when the dialog is shown."""
     finished: qtrio.Signal
     """The signal to be emitted when the dialog is finished."""
 
@@ -35,6 +36,27 @@ class DialogProtocol(Protocol):
     def teardown(self) -> None:
         """Hide and teardown the dialog."""
 
+
+check_basic_dialog_protocol = qtrio._util.ProtocolChecker[BasicDialogProtocol]()
+"""Assert proper implementation of :class:`qtrio.dialogs.BasicDialogProtocol` when type
+hint checking.
+
+Arguments:
+    cls: The class to verify.
+
+Returns:
+    The same class, unmodified, that was passed in.
+"""
+
+
+class DialogProtocol(BasicDialogProtocol, Protocol):
+    """The common interface used for working with QTrio dialogs.  To check that a class
+    implements this protocol see :func:`qtrio.dialogs.check_dialog_protocol`.
+    """
+
+    shown: qtrio.Signal
+    """The signal to be emitted when the dialog is shown."""
+
     async def wait(self) -> object:
         """Show the dialog, wait for the user interaction, and return the result.
 
@@ -44,24 +66,20 @@ class DialogProtocol(Protocol):
         """
 
 
-DialogProtocolT = typing.TypeVar("DialogProtocolT", bound=DialogProtocol)
+check_dialog_protocol = qtrio._util.ProtocolChecker[DialogProtocol]()
+"""Assert proper implementation of :class:`qtrio.dialogs.DialogProtocol` when type hint
+checking.
 
+Arguments:
+    cls: The class to verify.
 
-def check_dialog_protocol(
-    cls: typing.Type[DialogProtocolT],
-) -> typing.Type[DialogProtocolT]:
-    """Decorate a class with this to verify it implements the
-    :class:`qtrio.dialogs.DialogProtocol` when a type hint checker such as mypy is run
-    against the code.  At runtime the passed class is cleanly returned.
-
-    Arguments:
-        cls: The class to verify.
-    """
-    return cls
+Returns:
+    The same class, unmodified, that was passed in.
+"""
 
 
 @contextlib.contextmanager
-def _manage(dialog: DialogProtocol) -> typing.Generator[trio.Event, None, None]:
+def _manage(dialog: BasicDialogProtocol) -> typing.Generator[trio.Event, None, None]:
     """Manage the setup and teardown of a dialog including yielding a
     :class:`trio.Event` that is set when the dialog is finished.
 
@@ -120,10 +138,10 @@ class IntegerDialog:
     shown = qtrio.Signal(QtWidgets.QInputDialog)
     """See :attr:`qtrio.dialogs.DialogProtocol.shown`."""
     finished = qtrio.Signal(int)  # QtWidgets.QDialog.DialogCode
-    """See :attr:`qtrio.dialogs.DialogProtocol.finished`."""
+    """See :attr:`qtrio.dialogs.BasicDialogProtocol.finished`."""
 
     def setup(self) -> None:
-        """See :meth:`qtrio.dialogs.DialogProtocol.setup`."""
+        """See :meth:`qtrio.dialogs.BasicDialogProtocol.setup`."""
 
         self.result = None
 
@@ -143,7 +161,7 @@ class IntegerDialog:
         self.shown.emit(self.dialog)
 
     def teardown(self) -> None:
-        """See :meth:`qtrio.dialogs.DialogProtocol.teardown`."""
+        """See :meth:`qtrio.dialogs.BasicDialogProtocol.teardown`."""
 
         if self.dialog is not None:
             self.dialog.close()
@@ -216,10 +234,10 @@ class TextInputDialog:
     shown = qtrio.Signal(QtWidgets.QInputDialog)
     """See :attr:`qtrio.dialogs.DialogProtocol.shown`."""
     finished = qtrio.Signal(int)  # QtWidgets.QDialog.DialogCode
-    """See :attr:`qtrio.dialogs.DialogProtocol.finished`."""
+    """See :attr:`qtrio.dialogs.BasicDialogProtocol.finished`."""
 
     def setup(self) -> None:
-        """See :meth:`qtrio.dialogs.DialogProtocol.setup`."""
+        """See :meth:`qtrio.dialogs.BasicDialogProtocol.setup`."""
 
         self.result = None
 
@@ -242,7 +260,7 @@ class TextInputDialog:
         self.shown.emit(self.dialog)
 
     def teardown(self) -> None:
-        """See :meth:`qtrio.dialogs.DialogProtocol.teardown`."""
+        """See :meth:`qtrio.dialogs.BasicDialogProtocol.teardown`."""
 
         if self.dialog is not None:
             self.dialog.close()
@@ -321,6 +339,8 @@ class FileDialog:
     """The confirmation button."""
     reject_button: typing.Optional[QtWidgets.QPushButton] = None
     """The cancellation button."""
+    file_name_line_edit: typing.Optional[QtWidgets.QLineEdit] = None
+    """The file name line edit widget."""
 
     result: typing.Optional[trio.Path] = None
     """The path selected by the user."""
@@ -328,10 +348,29 @@ class FileDialog:
     shown = qtrio.Signal(QtWidgets.QFileDialog)
     """See :attr:`qtrio.dialogs.DialogProtocol.shown`."""
     finished = qtrio.Signal(int)  # QtWidgets.QDialog.DialogCode
-    """See :attr:`qtrio.dialogs.DialogProtocol.finished`."""
+    """See :attr:`qtrio.dialogs.BasicDialogProtocol.finished`."""
+
+    async def set_path(self, path: trio.Path) -> None:
+        """Set the directory and enter the file name in the text box.  Note that this
+        does not select the file in the file list.
+
+        Arguments:
+            path: The full path to the file to be set.
+        """
+        if self.dialog is None:
+            raise qtrio.DialogNotActiveError(
+                "File dialog not available for interaction."
+            )
+
+        self.dialog.setDirectory(os.fspath(path.parent))
+
+        if self.file_name_line_edit is None:  # pragma: no cover
+            raise qtrio.InternalError("File name line edit unexpectedly not known.")
+
+        self.file_name_line_edit.setText(path.name)
 
     def setup(self) -> None:
-        """See :meth:`qtrio.dialogs.DialogProtocol.setup`."""
+        """See :meth:`qtrio.dialogs.BasicDialogProtocol.setup`."""
 
         self.result = None
 
@@ -363,11 +402,12 @@ class FileDialog:
         buttons = _dialog_button_box_buttons_by_role(dialog=self.dialog)
         self.accept_button = buttons.get(QtWidgets.QDialogButtonBox.AcceptRole)
         self.reject_button = buttons.get(QtWidgets.QDialogButtonBox.RejectRole)
+        [self.file_name_line_edit] = self.dialog.findChildren(QtWidgets.QLineEdit)
 
         self.shown.emit(self.dialog)
 
     def teardown(self) -> None:
-        """See :meth:`qtrio.dialogs.DialogProtocol.teardown`."""
+        """See :meth:`qtrio.dialogs.BasicDialogProtocol.teardown`."""
 
         if self.dialog is not None:
             self.dialog.close()
@@ -375,6 +415,7 @@ class FileDialog:
         self.dialog = None
         self.accept_button = None
         self.reject_button = None
+        self.file_name_line_edit = None
 
     async def wait(self) -> trio.Path:
         """See :meth:`qtrio.dialogs.DialogProtocol.wait`."""
@@ -401,7 +442,7 @@ def create_file_save_dialog(
     default_file: typing.Optional[trio.Path] = None,
     options: QtWidgets.QFileDialog.Option = QtWidgets.QFileDialog.Option(),
 ) -> FileDialog:
-    """Create an open or save dialog.
+    """Create a file save dialog.
 
     Arguments:
         parent: See :attr:`qtrio.dialogs.FileDialog.parent`.
@@ -419,6 +460,30 @@ def create_file_save_dialog(
     )
 
 
+def create_file_open_dialog(
+    parent: typing.Optional[QtWidgets.QWidget] = None,
+    default_directory: typing.Optional[trio.Path] = None,
+    default_file: typing.Optional[trio.Path] = None,
+    options: QtWidgets.QFileDialog.Option = QtWidgets.QFileDialog.Option(),
+) -> FileDialog:
+    """Create a file open dialog.
+
+    Arguments:
+        parent: See :attr:`qtrio.dialogs.FileDialog.parent`.
+        default_directory: See :attr:`qtrio.dialogs.FileDialog.default_directory`.
+        default_file: See :attr:`qtrio.dialogs.FileDialog.default_file`.
+        options: See :attr:`qtrio.dialogs.FileDialog.options`.
+    """
+    return FileDialog(
+        parent=parent,
+        default_directory=default_directory,
+        default_file=default_file,
+        options=options,
+        file_mode=QtWidgets.QFileDialog.AnyFile,
+        accept_mode=QtWidgets.QFileDialog.AcceptOpen,
+    )
+
+
 @check_dialog_protocol
 @attr.s(auto_attribs=True)
 class MessageBox:
@@ -448,10 +513,10 @@ class MessageBox:
     shown = qtrio.Signal(QtWidgets.QMessageBox)
     """See :attr:`qtrio.dialogs.DialogProtocol.shown`."""
     finished = qtrio.Signal(int)  # QtWidgets.QDialog.DialogCode
-    """See :attr:`qtrio.dialogs.DialogProtocol.finished`."""
+    """See :attr:`qtrio.dialogs.BasicDialogProtocol.finished`."""
 
     def setup(self) -> None:
-        """See :meth:`qtrio.dialogs.DialogProtocol.setup`."""
+        """See :meth:`qtrio.dialogs.BasicDialogProtocol.setup`."""
 
         self.result = None
 
@@ -470,7 +535,7 @@ class MessageBox:
         self.shown.emit(self.dialog)
 
     def teardown(self) -> None:
-        """See :meth:`qtrio.dialogs.DialogProtocol.teardown`."""
+        """See :meth:`qtrio.dialogs.BasicDialogProtocol.teardown`."""
 
         if self.dialog is not None:
             self.dialog.close()
@@ -495,8 +560,8 @@ class MessageBox:
 
 
 def create_message_box(
-    title: str,
-    text: str,
+    title: str = "",
+    text: str = "",
     icon: QtWidgets.QMessageBox.Icon = QtWidgets.QMessageBox.Information,
     buttons: QtWidgets.QMessageBox.StandardButtons = QtWidgets.QMessageBox.Ok,
     parent: typing.Optional[QtWidgets.QWidget] = None,
@@ -511,3 +576,116 @@ def create_message_box(
         parent: See :attr:`qtrio.dialogs.MessageBox.parent`.
     """
     return MessageBox(icon=icon, title=title, text=text, buttons=buttons, parent=parent)
+
+
+@check_basic_dialog_protocol
+@attr.s(auto_attribs=True)
+class ProgressDialog:
+    """Manage a progress dialog for updating the user.  Generally instances should be
+    built via :func:`qtrio.dialogs.create_progress_dialog`.
+    """
+
+    title: str
+    """The message box title."""
+    text: str
+    """The message text shown as the progress bar label."""
+    cancel_button_text: typing.Optional[str]
+    """The cancel button text."""
+    minimum: int
+    """The progress value corresponding to no progress."""
+    maximum: int
+    """The progress value corresponding to completion."""
+
+    parent: typing.Optional[QtWidgets.QWidget] = None
+    """The parent widget for the dialog."""
+
+    dialog: typing.Optional[QtWidgets.QProgressDialog] = None
+    """The actual dialog widget instance."""
+    cancel_button: typing.Optional[QtWidgets.QPushButton] = None
+    """The cancellation button."""
+
+    shown = qtrio.Signal(QtWidgets.QMessageBox)
+    """See :attr:`qtrio.dialogs.DialogProtocol.shown`."""
+    finished = qtrio.Signal(int)  # QtWidgets.QDialog.DialogCode
+    """See :attr:`qtrio.dialogs.BasicDialogProtocol.finished`."""
+
+    def setup(self) -> None:
+        """See :meth:`qtrio.dialogs.BasicDialogProtocol.setup`."""
+
+        self.final_value = None
+
+        self.dialog = QtWidgets.QProgressDialog()
+        self.dialog.setWindowTitle(self.title)
+        self.dialog.setLabelText(self.text)
+        self.dialog.setMinimum(self.minimum)
+        self.dialog.setMaximum(self.maximum)
+        self.dialog.setParent(self.parent)
+
+        # TODO: adjust so we can use a context manager?
+        self.dialog.finished.connect(self.finished)
+
+        if self.cancel_button_text is not None:
+            self.dialog.setCancelButtonText(self.cancel_button_text)
+
+        self.dialog.show()
+
+        [self.cancel_button] = self.dialog.findChildren(QtWidgets.QPushButton)
+
+        self.shown.emit(self.dialog)
+
+    def teardown(self) -> None:
+        """See :meth:`qtrio.dialogs.BasicDialogProtocol.teardown`."""
+
+        if self.dialog is not None:
+            self.dialog.close()
+            self.dialog.finished.disconnect(self.finished)
+        self.dialog = None
+        self.cancel_button = None
+
+    @async_generator.asynccontextmanager
+    async def manage(self) -> typing.AsyncIterable[None]:
+        """A context manager to setup the progress dialog, cancel the managed context
+        and teardown the dialog when done.
+        """
+        with _manage(dialog=self):
+            if self.dialog is None:  # pragma: no cover
+                raise qtrio.InternalError(
+                    "Dialog not assigned while it is being managed."
+                )
+
+            with trio.CancelScope() as cancel_scope:
+                with qtrio._qt.connection(
+                    signal=self.dialog.canceled, slot=cancel_scope.cancel
+                ):
+                    yield
+
+            if self.dialog.wasCanceled():
+                raise qtrio.UserCancelledError()
+
+
+def create_progress_dialog(
+    title: str = "",
+    text: str = "",
+    cancel_button_text: typing.Optional[str] = None,
+    minimum: int = 0,
+    maximum: int = 0,
+    parent: typing.Optional[QtWidgets.QWidget] = None,
+) -> ProgressDialog:
+    """Create a progress dialog.
+
+    Arguments:
+        title: See :attr:`qtrio.dialogs.ProgressDialog.title`.
+        text: See :attr:`qtrio.dialogs.ProgressDialog.text`.
+        cancel_button_text: See :attr:`qtrio.dialogs.ProgressDialog.cancel_button_text`.
+        minimum: See :attr:`qtrio.dialogs.ProgressDialog.minimum`.
+        maximum: See :attr:`qtrio.dialogs.ProgressDialog.maximum`.
+        parent: See :attr:`qtrio.dialogs.ProgressDialog.parent`.
+    """
+    return ProgressDialog(
+        title=title,
+        text=text,
+        cancel_button_text=cancel_button_text,
+        minimum=minimum,
+        maximum=maximum,
+        parent=parent,
+    )
